angular.module("beamng.apps")
/**
 * Provides colour classes and some additional colour related functions
 */
.factory("TachoApps", ["TachoColours", "TachoBase", function(TachoColours, TachoBase) {
  /**
   * Represents a canvas object that can be drawn on
   */
  class Canvas {
    /**
     * Constructor
     * @param {HTMLDocument} document - the document object this canvas is on
     * @param {string} id - the canvas's identifier string
     */
    constructor(document, id) {
      this.element = document.getElementById(id);
      this.ID = id;

      this.setupCanvas();
    }

    /**
     * Ensures proper HD scaling of text to eliminate blurry text and geometry
     */
    setupCanvas() {
      // Ensure correct HD resolution on the canvas (fix for blurry text)
      this.element.width = this.element.getBoundingClientRect().width;
      this.element.height = this.element.getBoundingClientRect().height;
      this.get2DContext().imageSmoothingEnabled = true;
      this.show(false);
    }

    /**
     * Shows or hides the canvas
     * @param {boolean} show - whether to show or hide the canvas
     */
    show(show) {
      this.element.style.opacity = show ? "1" : "0";
    }

    /**
     * @returns {CanvasRenderingContext2D} the 2D context of the canvas
     */
    get2DContext() {
      return this.element.getContext("2d");
    }

    /**
     * Clears the canvas of all content drawn on it
     */
    clear() {
      let ctx = this.get2DContext();
      ctx.clearRect(0, 0, this.element.width, this.element.height);
    }

    /**
     * Sets a listener to a canvas for the left and right click action by a mouse
     * @param {function} leftFunc - the function to perform on a left click
     * @param {function} rightFunc - the function to perform on a right click
     */
    setupCanvasClick(leftFunc, rightFunc) {
      this.element.addEventListener("mousedown", function(event) {
        switch (event.button) {
          case 0:
            leftFunc(event);
            break;
          case 2:
            rightFunc(event);
            break;
        }
      });
    }

    /**
     * Resets the transformation of a canvas's 2D context
     * @param {CanvasRenderingContext2D} ctx - the context to reset the transformation of
     */
    static resetTransformationState(ctx) {
      ctx.setTransform(1, 0, 0, 1, 0, 0);
    }
  }

  /**
   * Abstract
   * Represents the base app that all modded apps will use/extend
   */
  class BaseApp {
    /**
     * Constructor
     * {@param bngApi} is required as it cannot be requested here by AngularJS's injector for some
     * reason :/
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {[String]} units - the list of units that can be used for this app
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, baseRadius, fontName, units, document, bngApi, UiUnits) {
      this.ID = id;
      this.FONT_NAME = fontName;
      this.UNITS = units;
      this.radius = baseRadius;
      this.CENTRE = new TachoBase.Vector(baseRadius, baseRadius);

      // Default to first
      this.unit = units[0];

      // A list of states the UI can be in
      this.states = Object.freeze({"preInit":1, "init":2, "loaded":3, "hidden":4});
      // The current state
      this.state = this.states.init;

      // The gauge object itself
      this.gauge = null;

      // The colour schemes available
      this.COLOUR_SCHEMES = TachoColours.COLOUR_SCHEMES;
      // The current colour scheme selected
      this.colourScheme = this.COLOUR_SCHEMES[0];

      // The foreground and background canvases
      this.CANVAS    = new Canvas(document, this.ID + "_canvas");
      this.BG_CANVAS = new Canvas(document, this.ID + "_bgCanvas");

      // The UiUnits service from BeamNG
      this.UiUnits = UiUnits;

      // Setup the background canvas's mouse action events
      this.initBGCanvas(bngApi);
    }

    /**
     * Sets up the background canvas, adding an event handler for the mouse action
     */
    initBGCanvas(bngApi) {
      let thisApp = this;

      this.CANVAS.setupCanvasClick(
          function(_event) {
            // Change unit
            thisApp.cycleUnit();
            thisApp.recreate(bngApi, true);
          },
          function(_event) {
            // Change colour scheme
            thisApp.cycleColourScheme();
            thisApp.recreate(bngApi, true);
          }
      );
    }

    /**
     * Recreates the gauge and re-draws the background canvas
     */
    recreate(bngApi, resave) {
      // Re-create gauge to reflect unit changes
      this.createGauge();
      // Show unit on background canvas
      this.drawBackgroundCanvas();

      if (resave)
          // Save this colour scheme and unit as default
        this.save(bngApi);
    }

    /**
     * Changes the current unit to the next available option
     */
    cycleUnit() {
      this.unit = BaseApp.cycleArray(this.unit, this.UNITS);
    }

    /**
     * Changes the current colour scheme to the next available option
     */
    cycleColourScheme() {
      this.colourScheme = BaseApp.cycleArray(this.colourScheme, this.COLOUR_SCHEMES);
    }

    /**
     * Saves the current colour scheme and unit to a file in the user's directory. Note that this
     * only saves the indexes with respect to the COLOUR_SCHEMES array and not the actual instance
     */
    save(bngApi) {
      let data = {
        unit: this.UNITS.indexOf(this.unit),
        colourScheme: this.COLOUR_SCHEMES.indexOf(this.colourScheme)
      }

      bngApi.engineLua("extensions.ui_tachometers.save('" + this.ID + "', " + bngApi.serializeToLua(data) + ")");
    }

    /**
     * Loads the settings file from the user's directory. If the file does not exist, this method
     * does nothing; else, it loads the unit and colour scheme to this app
     * @param {?} bngApi - the BeamNG API for interfacing with Lua
     */
    load(bngApi) {
      bngApi.engineLua("extensions.ui_tachometers.load('" + this.ID + "')", (response) => {
        if (!response)
          return;

        this.unit = this.UNITS[response.unit];
        this.colourScheme = this.COLOUR_SCHEMES[response.colourScheme];

        this.recreate(bngApi, false);
      });
    }

    /**
     * Draws to the main canvas
     */
    draw() {
      // this.prepCanvas(this.CANVAS);
      this.CANVAS.show(true);
      this.CANVAS.clear();
    }

    /**
     * Draws the app to the background canvas
     */
    drawBackgroundCanvas() {
      // this.prepCanvas(this.BG_CANVAS);
      this.BG_CANVAS.show(true);
      this.BG_CANVAS.clear();
    }

    /**
     * Hides the app by hiding both canvases
     */
    hide() {
      this.CANVAS.show(false);
      this.BG_CANVAS.show(false);
      this.state = this.states.hidden;
    }

    // prepCanvas(canvas) {
    //   let canvasElement = canvas.element;
    //   let dimensions = BaseApp.getObjectFitSize(
    //       canvasElement.clientWidth,
    //       canvasElement.clientHeight,
    //       canvasElement.width,
    //       canvasElement.height
    //   );
    //   canvasElement.width = dimensions.width;
    //   canvasElement.height = dimensions.height;
    //
    //   console.log(canvas, dimensions);
    //
    //   let ctx = canvas.get2DContext();
    //   let ratio = Math.min(
    //       canvasElement.clientWidth / this.radius,
    //       canvasElement.clientHeight / this.radius
    //   );
    //   ctx.scale(ratio, ratio);
    // }


    // adapted from: https://www.npmjs.com/package/intrinsic-scale
    static getObjectFitSize(containerWidth, containerHeight, width, height) {
      let doRatio = width / height;
      let cRatio = containerWidth / containerHeight;
      let test = doRatio > cRatio;

      let targetWidth;
      let targetHeight;

      if (test) {
        targetWidth = containerWidth;
        targetHeight = targetWidth / doRatio;
      } else {
        targetHeight = containerHeight;
        targetWidth = targetHeight * doRatio;
      }

      return {
        width: targetWidth,
        height: targetHeight
      };
    }

    /**
     * Returns the next value in an array, given a current value, looping around to the start if
     * the value is at the end of the array
     * @template T
     * @param {T} currValue - the current value
     * @param {[T]} array - the array containing a set of values, including {@param currValue}
     * @returns {T} the next value in {@param array}, succeeding {@param currValue}
     */
    static cycleArray(currValue, array) {
      // Find next index, looping back around to 0 if necessary
      let index = array.indexOf(currValue);
      index = (index+1) % array.length;

      return array[index];
    }

    /**
     * Resets the filter to prevent lingering filters
     */
    static resetFilter(ctx) {
      ctx.filter = "none";
    }

    /**
     * Reloads the UI, hiding it after 50ms. If instantReload is true, this is done
     * instantly
     * @param {BoostApp} app - the app to reload
     * @param {boolean} instantReload - whether to instantly reload instead of waiting
     */
    static reload(app, instantReload) {
      if (!app)
        return;

      let time = instantReload ? 0 : 50;
      setTimeout(() => {
        // It is unknown whether this vehicle will need this app, so hide it provisionally and
        // it is does, the app will re-show itself when updating occurs
        app.hide();
        app.state = app.states.init;
      }, time);
    }

    static changeSize(app, newWidth, newHeight) {
      // Chose smaller value to ensure fitting
      let radius = newWidth < newHeight ? newWidth : newHeight;

      app.CANVAS.element.width = radius;
      app.CANVAS.element.height = radius;

      app.BG_CANVAS.element.width = radius;
      app.BG_CANVAS.element.height = radius;

      console.log(radius, newWidth, newHeight);
    }


    /**
     * Creates the gauge element
     */
    createGauge() {
      console.error("createGauge() is abstract and must be implemented in the base class!");
    }
  }


  /**
   * Represents an vehicle's engine
   */
  class Engine {
    /**
     * Constructor
     */
    constructor() {
      // The current RPM
      this.currRPM = 0;
      // The maximum engine RPM
      this.maxRPM = 0;
      // The next thousand from the maximum RPM, treated as the absolute maximum RPM
      this.redLineRPM = 0;
      // The oil temperature
      this.engineTemp = 0;
    }


    setCurrRPM(rpm) {
      if (typeof rpm !== "undefined")
        this.currRPM = rpm;
    }

    setMaxRPM(maxRPM) {
      if (typeof maxRPM !== "undefined")
        this.maxRPM = maxRPM;
    }

    setRedLineRPM(redLineRPM) {
      if (typeof redLineRPM !== "undefined")
        this.redLineRPM = redLineRPM;
    }

    setEngineTemp(engineTemp) {
      if (typeof engineTemp !== "undefined")
        this.engineTemp = engineTemp;
    }

    getCurrentRPM() {
      return this.currRPM;
    }

    getMaxRPM() {
      return this.maxRPM;
    }

    getRedLineRPM() {
      return this.redLineRPM;
    }

    getTemperature() {
      return this.engineTemp;
    }


    /**
     * @returns {number} the ratio of current RPM to redline RPM
     */
    getCurrentRPMRatio() {
      return this.currRPM / this.redLineRPM;
    }

    /**
     * @returns {number} the ratio of maximum engine RPM and the redline (next thousand) RPM
     */
    getRedlineRPMRatio() {
      return this.maxRPM / this.redLineRPM;
    }

    /**
     * @returns {boolean} whether to shift up to the next gear
     */
    shouldShiftUp() {
      return this.currRPM >= this.maxRPM;
    }
  }


  /**
   * Abstract
   * Represents a base Tachometer modded app
   */
  class TachometerApp extends BaseApp {
    /**
     * Constructor
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {Range} rpmAngleRange - the RPM angle range of the gauge
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, rpmAngleRange, fontName, baseRadius, document, bngApi, UiUnits) {
      super(id, baseRadius, fontName, ["metric", "imperial"], document, bngApi, UiUnits);
      this.RPM_ANGLE_RANGE = rpmAngleRange;

      // The various attributes of the vehicle's state that may or may not be needed by the gauge
      this.lightsState = 0;
      this.currGearIndex = 0;
      this.speed = 0;

      this.engine = new Engine();

      this.pBrakeEngaged = false;
      this.fuel = 0;
    }

    /**
     * Updates the variables in the app's fields. All parameters are optional, and all are updated
     * via the respective field's setter function
     * @param {number} [currGearIndex] - the current gear's index the vehicle is in
     * @param {number} [speed] - the current speed of the vehicle
     * @param {number} [currRPM] - the current RPM of the engine
     * @param {number} [lightsState] - the state of the lights (0=off, 1=low, 2=high)
     * @param {boolean} [pBrakeEngaged] - whether the parking brake is engaged or not
     * @param {number} [engineTemp] - the temperature of the engine's water
     * @param {number} fuel - the percentage of fuel left
     */
    updateAllVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged, engineTemp, fuel) {
      this.setCurrGearIndex(currGearIndex);
      this.setSpeed(speed);
      this.setLightsState(lightsState);
      this.setParkingBrake(pBrakeEngaged);
      this.setFuel(fuel);

      this.engine.setCurrRPM(currRPM);
      this.engine.setEngineTemp(engineTemp);
    }

    /**
     * Updates the variables in the app's fields. All parameters are optional, and all are updated
     * via the respective field's setter function. If any variables have changed that affect the
     * background canvas, the flag needsRedraw is set to true.
     * @param {number} [currGearIndex] - the current gear's index the vehicle is in
     * @param {number} [speed] - the current speed of the vehicle
     * @param {number} [currRPM] - the current RPM of the engine
     * @param {number} [lightsState] - the state of the lights (0=off, 1=low, 2=high)
     * @param {boolean} [pBrakeEngaged] - whether the parking brake is engaged or not
     * @param {number} [engineTemp] - the temperature of the engine's water
     * @param {number} fuel - the percentage of fuel left
     */
    updateVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged, engineTemp, fuel) {
      let prevState = this.collateBGVars();
      this.updateAllVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged, engineTemp, fuel);
      let newState = this.collateBGVars();

      // If any variables for the background canvas have changed, the background canvas
      // requires a re-draw
      if (!prevState.every(function(v, i) {return v === newState[i]}))
        this.needsRedraw = true;
    }

    /**
     * Initialises the app's fields
     * @param {number} currGearIndex - the current gear's index the vehicle is in
     * @param {number} speed - the current speed of the vehicle
     * @param {number} currRPM - the current RPM of the engine
     * @param {number} maxRPM - the maximum RPM of the engine
     * @param {number} lightsState - the state of the lights (0=off, 1=low, 2=high)
     * @param {number} fuel - the percentage of fuel left
     */
    initFields(currGearIndex, speed, currRPM, maxRPM, lightsState, fuel) {
      this.updateVariables(currGearIndex, speed, currRPM, lightsState, fuel);
      this.engine.setMaxRPM(maxRPM);
      // Round to next thousand
      this.engine.setRedLineRPM(Math.floor(maxRPM / 1000 + 1) * 1000);

      this.createGauge();
    }

    /**
     * @returns {[*]} the list of all variables that are used in the drawing on the background
     * canvas. This includes the gear index, the lights' state and the parking brake
     */
    collateBGVars() {
      return [this.currGearIndex, this.lightsState, this.pBrakeEngaged];
    }

    setCurrGearIndex(gearIndex) {
      if (typeof gearIndex !== "undefined")
        this.currGearIndex = gearIndex;
    }

    setSpeed(speed) {
      if (typeof speed !== "undefined")
        this.speed = speed;
    }

    setLightsState(lightsState) {
      if (typeof lightsState !== "undefined" && lightsState >= 0 && lightsState <= 2)
        this.lightsState = lightsState;
    }

    setParkingBrake(pBrakeEngaged) {
      if (typeof pBrakeEngaged !== "undefined")
        this.pBrakeEngaged = pBrakeEngaged;
    }

    setFuel(fuel) {
      if (typeof fuel !== "undefined")
        this.fuel = fuel;
    }

    /**
     * @returns {number} the angle at which redline RPM occurs on the RPM gauge
     */
    getRedlineRPMAngle() {
      return this.RPM_ANGLE_RANGE.getLinearRatio(this.engine.getRedlineRPMRatio());
    }

    /**
     * Converts a gear index into its string name.
     * Mapping:
     *   < -1  -> R2...R(|n|+1)
     *   = -1  -> R
     *   =  0  -> N
     *  >=  1  -> 1...n
     * @param {number} gearIndex - the gear's index
     * @returns {string} the string name for that gear index
     */
    static gearIndexToName(gearIndex) {
      let gearName;

      if (gearIndex === 0) {
        gearName = "N";
      } else if (gearIndex < 0) {
        // Add index of reverse gear if more than one exists
        if (gearIndex < -1)
          gearName = "R" + Math.abs(gearIndex);
        else
          gearName = "R";
      } else {
        gearName = gearIndex.toString();
      }

      return gearName;
    }
  }

  /**
   * Abstract
   * Represents a base Boost modded app
   */
  class BoostApp extends BaseApp {
    /**
     * Constructor
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {Range} boostAngleRange - the boost angle range of the gauge
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, boostAngleRange, fontName, baseRadius, document, bngApi, UiUnits) {
      super(id, baseRadius, fontName, ["inHg", "bar", "psi", "kPa"], document, bngApi, UiUnits);
      this.BOOST_ANGLE_RANGE = boostAngleRange;

      this.maxBoost = 0;
      this.currBoost = 0;
    }

    /**
     * Initialises the app's fields.
     * Note that both values should be their 'raw' values, so that they can be converted to specific
     * units later
     * @param {number} currBoost - the current boost of the turbo
     * @param {number} maxBoost - the maximum possible boost value
     */
    initFields(currBoost, maxBoost) {
      this.setMaxBoost(maxBoost);
      this.setCurrBoost(currBoost);

      this.createGauge();
    }

    setMaxBoost(maxBoost) {
      if (typeof maxBoost !== "undefined")
        this.maxBoost = maxBoost;
    }

    setCurrBoost(currBoost) {
      if (typeof currBoost !== "undefined")
        this.currBoost = currBoost;
    }
  }

  this.Canvas = Canvas;
  this.BaseApp = BaseApp;
  this.TachometerApp = TachometerApp;
  this.BoostApp = BoostApp;

  return this;
}])